"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const fast_copy_1 = __importDefault(require("fast-copy"));
const process_1 = require("process");
const healthCache_1 = require("../../src/accounts/healthCache");
const mangoAccount_1 = require("../../src/accounts/mangoAccount");
const client_1 = require("../../src/client");
const constants_1 = require("../../src/constants");
const I80F48_1 = require("../../src/numbers/I80F48");
const utils_1 = require("../../src/utils");
const CLUSTER_URL = process.env.CLUSTER_URL_OVERRIDE || process.env.MB_CLUSTER_URL;
const USER_KEYPAIR = process.env.USER_KEYPAIR_OVERRIDE || process.env.MB_PAYER_KEYPAIR;
const MANGO_ACCOUNT_PK = new web3_js_1.PublicKey(process.env.MANGO_ACCOUNT_PK || web3_js_1.PublicKey.default.toBase58());
const CLUSTER = process.env.CLUSTER_OVERRIDE || 'mainnet-beta';
async function debugUser(client, group, mangoAccount) {
    console.log(mangoAccount.toString(group));
    await mangoAccount.reload(client);
    console.log('mangoAccount.getEquity() ' +
        (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getEquity(group).toNumber()));
    console.log('mangoAccount.getHealth(HealthType.init) ' +
        (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getHealth(group, mangoAccount_1.HealthType.init).toNumber()));
    console.log('HealthCache.fromMangoAccount(group,mangoAccount).health(HealthType.init) ' +
        (0, utils_1.toUiDecimalsForQuote)(healthCache_1.HealthCache.fromMangoAccount(group, mangoAccount)
            .health(mangoAccount_1.HealthType.init)
            .toNumber()));
    console.log('mangoAccount.getHealthRatio(HealthType.init) ' +
        mangoAccount.getHealthRatio(group, mangoAccount_1.HealthType.init).toNumber());
    console.log('mangoAccount.getHealthRatioUi(HealthType.init) ' +
        mangoAccount.getHealthRatioUi(group, mangoAccount_1.HealthType.init));
    console.log('mangoAccount.getHealthRatio(HealthType.maint) ' +
        mangoAccount.getHealthRatio(group, mangoAccount_1.HealthType.maint).toNumber());
    console.log('mangoAccount.getHealthRatioUi(HealthType.maint) ' +
        mangoAccount.getHealthRatioUi(group, mangoAccount_1.HealthType.maint));
    console.log('mangoAccount.getCollateralValue() ' +
        (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getCollateralValue(group).toNumber()));
    console.log('mangoAccount.getAssetsValue() ' +
        (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getAssetsValue(group).toNumber()));
    console.log('mangoAccount.getLiabsValue() ' +
        (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getLiabsValue(group).toNumber()));
    async function getMaxWithdrawWithBorrowForTokenUiWrapper(token) {
        console.log(`mangoAccount.getMaxWithdrawWithBorrowForTokenUi(group, ${token}) ` +
            mangoAccount.getMaxWithdrawWithBorrowForTokenUi(group, group.banksMapByName.get(token)[0].mint));
    }
    for (const srcToken of Array.from(group.banksMapByName.keys())) {
        await getMaxWithdrawWithBorrowForTokenUiWrapper(srcToken);
    }
    function getMaxSourceForTokenSwapWrapper(src, tgt) {
        // Turn on for debugging specific pairs
        // if (src != 'USDC' || tgt != 'MNGO') return;
        let maxSourceUi;
        try {
            maxSourceUi = mangoAccount.getMaxSourceUiForTokenSwap(group, group.banksMapByName.get(src)[0].mint, group.banksMapByName.get(tgt)[0].mint);
        }
        catch (error) {
            console.log(`Error for ${src}->${tgt}, ` + error.toString());
        }
        const maxTargetUi = maxSourceUi *
            (group.banksMapByName.get(src)[0].uiPrice /
                group.banksMapByName.get(tgt)[0].uiPrice);
        const sim = mangoAccount.simHealthRatioWithTokenPositionUiChanges(group, [
            {
                mintPk: group.banksMapByName.get(src)[0].mint,
                uiTokenAmount: -maxSourceUi,
            },
            {
                mintPk: group.banksMapByName.get(tgt)[0].mint,
                uiTokenAmount: maxTargetUi,
            },
        ]);
        console.log(`getMaxSourceForTokenSwap ${src.padEnd(4)} ${tgt.padEnd(4)} ` +
            maxSourceUi.toFixed(3).padStart(10) +
            `, health ratio after (${sim.toFixed(3).padStart(10)})`);
    }
    for (const srcToken of Array.from(group.banksMapByName.keys()).sort()) {
        for (const tgtToken of Array.from(group.banksMapByName.keys()).sort()) {
            getMaxSourceForTokenSwapWrapper(srcToken, tgtToken);
        }
    }
    function getMaxForPerpWrapper(perpMarket) {
        const maxQuoteUi = mangoAccount.getMaxQuoteForPerpBidUi(group, perpMarket.perpMarketIndex);
        const simMaxQuote = mangoAccount.simHealthRatioWithPerpBidUiChanges(group, perpMarket.perpMarketIndex, maxQuoteUi / perpMarket.uiPrice);
        const maxBaseUi = mangoAccount.getMaxBaseForPerpAskUi(group, perpMarket.perpMarketIndex);
        const simMaxBase = mangoAccount.simHealthRatioWithPerpAskUiChanges(group, perpMarket.perpMarketIndex, maxBaseUi);
        console.log(`getMaxPerp ${perpMarket.name.padStart(10)} getMaxQuoteForPerpBidUi ${maxQuoteUi
            .toFixed(3)
            .padStart(10)} health ratio after (${simMaxQuote
            .toFixed(3)
            .padStart(10)}), getMaxBaseForPerpAskUi ${maxBaseUi
            .toFixed(3)
            .padStart(10)} health ratio after (${simMaxBase
            .toFixed(3)
            .padStart(10)})`);
    }
    for (const perpMarket of Array.from(group.perpMarketsMapByMarketIndex.values())) {
        getMaxForPerpWrapper(perpMarket);
    }
    function getMaxForSerum3Wrapper(serum3Market) {
        console.log(`getMaxQuoteForSerum3BidUi ${serum3Market.name} ` +
            mangoAccount.getMaxQuoteForSerum3BidUi(group, serum3Market.serumMarketExternal));
        console.log(`- simHealthRatioWithSerum3BidUiChanges  ${serum3Market.name} ` +
            mangoAccount.simHealthRatioWithSerum3BidUiChanges(group, mangoAccount.getMaxQuoteForSerum3BidUi(group, serum3Market.serumMarketExternal), serum3Market.serumMarketExternal, mangoAccount_1.HealthType.init));
        console.log(`getMaxBaseForSerum3AskUi ${serum3Market.name} ` +
            mangoAccount.getMaxBaseForSerum3AskUi(group, serum3Market.serumMarketExternal));
        console.log(`- simHealthRatioWithSerum3BidUiChanges  ${serum3Market.name} ` +
            mangoAccount.simHealthRatioWithSerum3AskUiChanges(group, mangoAccount.getMaxBaseForSerum3AskUi(group, serum3Market.serumMarketExternal), serum3Market.serumMarketExternal, mangoAccount_1.HealthType.init));
    }
    for (const serum3Market of Array.from(group.serum3MarketsMapByExternal.values())) {
        getMaxForSerum3Wrapper(serum3Market);
    }
    // Liquidation price for perp positions
    for (const pp of mangoAccount.perpActive()) {
        const pm = group.getPerpMarketByMarketIndex(pp.marketIndex);
        const health = (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getHealth(group, mangoAccount_1.HealthType.maint));
        if (
        // pp.getNotionalValueUi(pm) > 1000 &&
        // !(pp.getNotionalValueUi(pm) < health && pp.getBasePosition(pm).isPos())
        // eslint-disable-next-line no-constant-condition
        true) {
            const then = Date.now();
            const startUsage = (0, process_1.cpuUsage)();
            const lp = await pp.getLiquidationPrice(group, mangoAccount);
            if (lp == null || lp.lt((0, I80F48_1.ZERO_I80F48)())) {
                continue;
            }
            const lpUi = group
                .getPerpMarketByMarketIndex(pp.marketIndex)
                .priceNativeToUi(lp.toNumber());
            const gClone = (0, fast_copy_1.default)(group);
            gClone.getPerpMarketByMarketIndex(pm.perpMarketIndex)._price = lp;
            const simHealth = (0, utils_1.toUiDecimalsForQuote)(mangoAccount.getHealth(gClone, mangoAccount_1.HealthType.maint));
            const now = Date.now();
            const endUsage = (0, process_1.cpuUsage)(startUsage);
            console.log(` - ${pm.name}, health: ${health.toLocaleString()}, side: ${pp.getBasePosition(pm).isPos() ? 'LONG' : 'SHORT'}, notional: ${pp
                .getNotionalValueUi(pm)
                .toLocaleString()}, liq price: ${lpUi.toLocaleString()}, sim health: ${simHealth.toLocaleString()}, time ${now - then}ms, cpu usage ${(endUsage['user'] / 1000).toLocaleString()}ms`);
        }
    }
}
async function main() {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(CLUSTER_URL, options);
    const wallet = new anchor_1.Wallet(new web3_js_1.Keypair());
    const provider = new anchor_1.AnchorProvider(connection, wallet, options);
    const client = client_1.MangoClient.connect(provider, CLUSTER, constants_1.MANGO_V4_ID[CLUSTER], {
        idsSource: 'api',
    });
    const group = await client.getGroup(new web3_js_1.PublicKey('78b8f4cGCwmZ9ysPFMWLaLTkkaYnUjwMJYStWe5RTSSX'));
    const mangoAccounts = await client.getAllMangoAccounts(group, true);
    mangoAccounts.sort((a, b) => b.getEquity(group).cmp(a.getEquity(group)));
    for (const mangoAccount of mangoAccounts) {
        if (mangoAccount.publicKey.equals(new web3_js_1.PublicKey(MANGO_ACCOUNT_PK))) {
            console.log(`account https://app.mango.markets/?address=${mangoAccount.publicKey}`);
            await debugUser(client, group, mangoAccount);
        }
    }
    process.exit();
}
try {
    main();
}
catch (error) {
    console.log(error);
}
