"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const listingTools_1 = require("@blockworks-foundation/mango-v4-settings/lib/helpers/listingTools");
const anchor_1 = require("@coral-xyz/anchor");
const anchor_2 = require("@project-serum/anchor");
const spl_governance_1 = require("@solana/spl-governance");
const builder_1 = require("../src/builder");
const web3_js_1 = require("@solana/web3.js");
const fs_1 = __importDefault(require("fs"));
const client_1 = require("../src/client");
const clientIxParamBuilder_1 = require("../src/clientIxParamBuilder");
const constants_1 = require("../src/constants");
const I80F48_1 = require("../src/numbers/I80F48");
const risk_1 = require("../src/risk");
const utils_1 = require("../src/utils");
const constants_2 = require("./governanceInstructions/constants");
const createProposal_1 = require("./governanceInstructions/createProposal");
const voteStakeRegistryClient_1 = require("./governanceInstructions/voteStakeRegistryClient");
const { MB_CLUSTER_URL, PROPOSAL_TITLE, PROPOSAL_LINK, VSR_DELEGATE_KEYPAIR, VSR_DELEGATE_FROM_PK, DRY_RUN, } = process.env;
const getApiTokenName = (bankName) => {
    if (bankName === 'ETH (Portal)') {
        return 'ETH';
    }
    return bankName;
};
async function buildClient() {
    return await client_1.MangoClient.connectDefault(MB_CLUSTER_URL);
}
async function setupWallet() {
    const clientKeypair = web3_js_1.Keypair.fromSecretKey(Buffer.from(JSON.parse(fs_1.default.readFileSync(VSR_DELEGATE_KEYPAIR, 'utf-8'))));
    const clientWallet = new anchor_1.Wallet(clientKeypair);
    return clientWallet;
}
async function setupVsr(connection, clientWallet) {
    const options = anchor_1.AnchorProvider.defaultOptions();
    const provider = new anchor_1.AnchorProvider(connection, clientWallet, options);
    const vsrClient = await voteStakeRegistryClient_1.VsrClient.connect(provider, voteStakeRegistryClient_1.DEFAULT_VSR_ID);
    return vsrClient;
}
async function getTotalLiqorEquity(client, group, mangoAccounts) {
    const liqors = (await (await (await (0, utils_1.buildFetch)())(`https://api.mngo.cloud/data/v4/stats/liqors-over_period?over_period=1MONTH`, {
        mode: 'cors',
        headers: {
            'Content-Type': 'application/json',
            'Access-Control-Allow-Origin': '*',
        },
    })).json()).map((data) => new web3_js_1.PublicKey(data['liqor']));
    const ttlLiqorEquity = (await (0, risk_1.getEquityForMangoAccounts)(client, group, liqors, mangoAccounts)).reduce((partialSum, ae) => partialSum + ae.Equity.val, 0);
    return ttlLiqorEquity;
}
function getPriceImpactForBank(midPriceImpacts, bank, priceImpactPercent = 1) {
    const tokenToPriceImpact = midPriceImpacts
        .filter((x) => x.avg_price_impact_percent < priceImpactPercent)
        .reduce((acc, val) => {
        if (!acc[val.symbol] ||
            val.target_amount > acc[val.symbol].target_amount) {
            acc[val.symbol] = val;
        }
        return acc;
    }, {});
    const priceImpact = tokenToPriceImpact[getApiTokenName(bank.name)];
    return priceImpact;
}
async function updateTokenParams() {
    const [client, wallet] = await Promise.all([buildClient(), setupWallet()]);
    const vsrClient = await setupVsr(client.connection, wallet);
    const group = await client.getGroup(constants_1.MANGO_V4_MAIN_GROUP);
    const instructions = [];
    const allMangoAccounts = await client.getAllMangoAccounts(group, true);
    const stepSize = 1;
    const ttlLiqorEquityUi = await getTotalLiqorEquity(client, group, allMangoAccounts);
    const midPriceImpacts = (0, listingTools_1.getMidPriceImpacts)(group.pis);
    const pisForLiqor = [];
    // eslint-disable-next-line no-constant-condition
    if (false) {
        const pisForLiqor = await (0, risk_1.buildGroupGrid)(group, allMangoAccounts, stepSize);
    }
    // eslint-disable-next-line no-constant-condition
    if (false) {
        // Deposit limits header
        console.log(`${'name'.padStart(20)} ${'maxLiqBatchUi'.padStart(15)} ${'maxLiqBatchUi'.padStart(15)} ${'sellImpact'.padStart(12)}$ ${'pi %'.padStart(12)}% ${'aNDUi'.padStart(20)}${'aNDQuoteUi'.padStart(20)} ${'uiDeposits'.padStart(12)} ${'uiDeposits'.padStart(12)} ${'depositLimitsUi'.padStart(12)}`);
    }
    Array.from(group.banksMapByTokenIndex.values())
        .map((banks) => banks[0])
        .sort((a, b) => a.name.localeCompare(b.name))
        .forEach(async (bank) => {
        const builder = (0, builder_1.Builder)(clientIxParamBuilder_1.NullTokenEditParams);
        let change = false;
        // eslint-disable-next-line no-constant-condition
        if (true) {
            const tier = Object.values(listingTools_1.LISTING_PRESETS).find((x) => x.initLiabWeight.toFixed(1) === '1.8'
                ? x.initLiabWeight.toFixed(1) ===
                    bank?.initLiabWeight.toNumber().toFixed(1) &&
                    x.reduceOnly === bank.reduceOnly
                : x.initLiabWeight.toFixed(1) ===
                    bank?.initLiabWeight.toNumber().toFixed(1));
            if (!tier) {
                console.log(`${bank.name}, no tier found`);
            }
            else {
                console.log(`${bank.name.padStart(10)}, ${bank.loanFeeRate
                    .mul(I80F48_1.I80F48.fromNumber(100))
                    .toFixed(2)}, ${bank.loanOriginationFeeRate
                    .mul(I80F48_1.I80F48.fromNumber(100))
                    .toFixed(2)}, ${tier?.preset_name.padStart(5)}, ${(tier.loanFeeRate * 100).toFixed(2)}, ${(tier.loanOriginationFeeRate * 100).toFixed(2)}`);
                builder.loanFeeRate(tier.loanFeeRate);
                builder.loanOriginationFeeRate(tier.loanOriginationFeeRate);
                builder.flashLoanSwapFeeRate(tier.loanOriginationFeeRate);
                change = true;
            }
        }
        try {
            // formulas are sourced from here
            // https://www.notion.so/mango-markets/Mango-v4-Risk-parameter-recommendations-d309cdf5faac4aeea7560356e68532ab
            // const priceImpact = getPriceImpactForBank(midPriceImpacts, bank);
            // const scaleStartQuoteUi = Math.min(
            //   50 * ttlLiqorEquityUi,
            //   4 * priceImpact.target_amount,
            // );
            if (!bank.areBorrowsReduceOnly()) {
                // eslint-disable-next-line no-constant-condition
                if (false) {
                    // Net borrow limits
                    const netBorrowLimitPerWindowQuote = Math.max(10_000, Math.min(bank.uiDeposits() * bank.uiPrice, 300_000) / 3 +
                        Math.max(0, bank.uiDeposits() * bank.uiPrice - 300_000) / 5);
                    builder.netBorrowLimitPerWindowQuote((0, utils_1.toNativeI80F48ForQuote)(netBorrowLimitPerWindowQuote).toNumber());
                    change = true;
                    if (netBorrowLimitPerWindowQuote !=
                        (0, utils_1.toUiDecimalsForQuote)(bank.netBorrowLimitPerWindowQuote)) {
                        console.log(`${bank.name} new - ${netBorrowLimitPerWindowQuote.toLocaleString()}, old - ${(0, utils_1.toUiDecimalsForQuote)(bank.netBorrowLimitPerWindowQuote).toLocaleString()}`);
                    }
                }
                // Deposit limits
                // eslint-disable-next-line no-constant-condition
                if (false) {
                    if (bank.maintAssetWeight.toNumber() > 0) {
                        {
                            // Find asset's largest batch in $ we would need to liquidate, batches are extreme points of a range of price drop,
                            // range is constrained by leverage provided
                            // i.e. how much volatility we expect
                            const r = (0, risk_1.findLargestAssetBatchUi)(pisForLiqor, bank.name, Math.round(bank.maintAssetWeight.toNumber() * 100), 100 - Math.round(bank.maintAssetWeight.toNumber() * 100), stepSize);
                            const maxLiqBatchQuoteUi = r[0];
                            const maxLiqBatchUi = r[1];
                            const sellImpact = getPriceImpactForBank(midPriceImpacts, bank, (bank.liquidationFee.toNumber() * 100) / 2);
                            // Deposit limit = sell impact - largest batch
                            const allowedNewDepositsQuoteUi = sellImpact.target_amount - maxLiqBatchQuoteUi;
                            const allowedNewDepositsUi = sellImpact.target_amount / bank.uiPrice -
                                maxLiqBatchQuoteUi / bank.uiPrice;
                            const depositLimitUi = bank.uiDeposits() + allowedNewDepositsUi;
                            // LOG
                            // console.log(
                            //   `${bank.name.padStart(20)} ${maxLiqBatchUi
                            //     .toLocaleString()
                            //     .padStart(15)} ${maxLiqBatchQuoteUi
                            //     .toLocaleString()
                            //     .padStart(15)}$ ${sellImpact.target_amount
                            //     .toLocaleString()
                            //     .padStart(12)}$ ${sellImpact.avg_price_impact_percent
                            //     .toLocaleString()
                            //     .padStart(12)}% ${allowedNewDepositsUi
                            //     .toLocaleString()
                            //     .padStart(20)}${allowedNewDepositsQuoteUi
                            //     .toLocaleString()
                            //     .padStart(20)}$ ${bank
                            //     .uiDeposits()
                            //     .toLocaleString()
                            //     .padStart(12)} ${(bank.uiDeposits() * bank.uiPrice)
                            //     .toLocaleString()
                            //     .padStart(12)}$ ${depositLimitUi
                            //     .toLocaleString()
                            //     .padStart(12)}`,
                            // );
                            builder.depositLimit((0, utils_1.toNative)(depositLimitUi, bank.mintDecimals));
                            change = true;
                        }
                    }
                }
                const params = builder.build();
                console.log(bank.name);
                console.log(params.loanFeeRate);
                console.log(params.loanOriginationFeeRate);
                console.log(params.flashLoanSwapFeeRate);
                const ix = await client.program.methods
                    .tokenEdit(params.oracle, params.oracleConfig, params.groupInsuranceFund, params.interestRateParams, params.loanFeeRate, params.loanOriginationFeeRate, params.maintAssetWeight, params.initAssetWeight, params.maintLiabWeight, params.initLiabWeight, params.liquidationFee, params.stablePriceDelayIntervalSeconds, params.stablePriceDelayGrowthLimit, params.stablePriceGrowthLimit, params.minVaultToDepositsRatio, params.netBorrowLimitPerWindowQuote !== null
                    ? new anchor_2.BN(params.netBorrowLimitPerWindowQuote)
                    : null, params.netBorrowLimitWindowSizeTs !== null
                    ? new anchor_2.BN(params.netBorrowLimitWindowSizeTs)
                    : null, params.borrowWeightScaleStartQuote, params.depositWeightScaleStartQuote, params.resetStablePrice ?? false, params.resetNetBorrowLimit ?? false, params.reduceOnly, params.name, params.forceClose, params.tokenConditionalSwapTakerFeeRate, params.tokenConditionalSwapMakerFeeRate, params.flashLoanSwapFeeRate, params.interestCurveScaling, params.interestTargetUtilization, params.maintWeightShiftStart, params.maintWeightShiftEnd, params.maintWeightShiftAssetTarget, params.maintWeightShiftLiabTarget, params.maintWeightShiftAbort ?? false, false, // setFallbackOracle, unused
                params.depositLimit, params.zeroUtilRate, params.platformLiquidationFee, params.disableAssetLiquidation, params.collateralFeePerDay, params.forceWithdraw)
                    .accounts({
                    group: group.publicKey,
                    oracle: bank.oracle,
                    admin: group.admin,
                    mintInfo: group.mintInfosMapByTokenIndex.get(bank.tokenIndex)
                        ?.publicKey,
                })
                    .remainingAccounts([
                    {
                        pubkey: bank.publicKey,
                        isWritable: true,
                        isSigner: false,
                    },
                ])
                    .instruction();
                const tx = new web3_js_1.Transaction({ feePayer: wallet.publicKey }).add(ix);
                const simulated = await client.connection.simulateTransaction(tx);
                if (simulated.value.err) {
                    console.log('error', simulated.value.logs);
                    throw simulated.value.logs;
                }
                if (change) {
                    instructions.push(ix);
                }
            }
        }
        catch (error) {
            console.log(`....Skipping ${bank.name}, ${error}`);
        }
    });
    const tokenOwnerRecordPk = await (0, spl_governance_1.getTokenOwnerRecordAddress)(constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK, constants_2.MANGO_MINT, new web3_js_1.PublicKey(VSR_DELEGATE_FROM_PK));
    const [tokenOwnerRecord, proposals] = await Promise.all([
        (0, spl_governance_1.getTokenOwnerRecord)(client.connection, tokenOwnerRecordPk),
        (0, spl_governance_1.getAllProposals)(client.connection, constants_2.MANGO_GOVERNANCE_PROGRAM, constants_2.MANGO_REALM_PK),
    ]);
    const walletSigner = wallet;
    console.log(DRY_RUN);
    if (!DRY_RUN) {
        const proposalAddress = await (0, createProposal_1.createProposal)(client.connection, walletSigner, constants_2.MANGO_DAO_WALLET_GOVERNANCE, tokenOwnerRecord, PROPOSAL_TITLE
            ? PROPOSAL_TITLE
            : 'Update loan fee, loan origination fee, and flash loan fees in mango-v4', PROPOSAL_LINK ?? '', Object.values(proposals).length, instructions, vsrClient, false);
        console.log(proposalAddress.toBase58());
    }
}
async function main() {
    try {
        await updateTokenParams();
    }
    catch (error) {
        console.log(error);
    }
}
try {
    main();
}
catch (error) {
    console.log(error);
}
